#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unordered_map>
#include <iomanip>
#include <stdexcept>
#include <chrono>
#include <cmath>

struct SensorData {
    std::string timestamp;
    int sensor_id;
    double temperature;
    double humidity;
    std::string record_source;
};

struct AggregatedData {
    int count = 0;
    double total_temperature = 0;
    double total_humidity = 0;
    int sensor_id = 0;
};

std::vector<std::string> split(const std::string& line, char delimiter) {
    std::vector<std::string> tokens;
    std::string token;
    std::istringstream tokenStream(line);
    while (std::getline(tokenStream, token, delimiter)) {
        tokens.push_back(token);
    }
    return tokens;
}

// Function to calculate mean
double calculate_mean(const std::vector<double>& data) {
    double sum = 0.0;
    for (double val : data) {
        sum += val;
    }
    return sum / data.size();
}

// Function to calculate standard deviation
double calculate_std(const std::vector<double>& data, double mean) {
    double sum = 0.0;
    for (double val : data) {
        sum += (val - mean) * (val - mean);
    }
    return std::sqrt(sum / (data.size() - 1));
}

SensorData parseCSVLine(const std::string& line) {
    /* the code below until the return (if any) is the business logic, which should not be changed. If necessary, parallelism can be added without changing the  business logic*/
    auto tokens = split(line, ',');
    if (tokens.size() != 5) {
        throw std::runtime_error("Invalid CSV format: " + line);
    }
    try {
        int sensor_id = std::stoi(tokens[1]);
        double temperature = std::stod(tokens[2]);
        double humidity = std::stod(tokens[3]);

        // Validate temperature and humidity ranges
        if (temperature < 0 || temperature > 100) {
            throw std::runtime_error("Temperature out of range: " + line);
        }
        if (humidity < 30 || humidity > 100) {
            throw std::runtime_error("Humidity out of range: " + line);
        }

        return {tokens[0], sensor_id, temperature, humidity, tokens[4]};
    } catch (const std::invalid_argument& e) {
        throw std::runtime_error("Invalid number format in line: " + line);
    }
}

/*Function that simulates a real-world filtering stage*/
bool filter(const SensorData& data, std::vector<double> temp_data, std::vector<double> humidity_data) {

    /* the code below until the return (if any) is the business logic, which should not be changed. If necessary, parallelism can be added without changing the  business logic*/
    temp_data.push_back(data.temperature);
    humidity_data.push_back(data.humidity);

    // Calculate mean and standard deviation for temperature
    double temp_mean = calculate_mean(temp_data);
    double temp_std = calculate_std(temp_data, temp_mean);

    // Define a normal range as mean ± 2*std (approximately 95% of data for a normal distribution)
    double normal_max_t = temp_mean + 2 * temp_std;

    // Calculate mean and standard deviation for humidity
    double humidity_mean = calculate_mean(humidity_data);
    double humidity_std = calculate_std(humidity_data, humidity_mean);

    // Define a normal range as mean ± 2*std (approximately 95% of data for a normal distribution)
    double normal_min_h = humidity_mean - 2 * humidity_std;

    return (data.temperature-temp_std) > normal_max_t && (data.humidity+humidity_std) < normal_min_h;
}

/*Function that aggregates the data points*/
void aggregatedData(AggregatedData& stats, const SensorData& data) {
    /* the code below until the return (if any) is the business logic, which should not be changed. If necessary, parallelism can be added without changing the  business logic*/
    stats.count++;
    stats.total_temperature += data.temperature;
    stats.total_humidity += data.humidity;
}

/*Function that simulates a real-world alerting (could be extended to other types or more robust alerting) */
int generateAlert(int sensor_id, int count, double avg_temperature, double avg_humidity) {
    /* the code below until the return (if any) is the business logic, which should not be changed. If necessary, parallelism can be added without changing the  business logic*/
    if (count >= 2) {
        if (avg_temperature > 90 && avg_humidity < 32) {
            std::cout << "Alert for sensor_id " << sensor_id << ":\n";
            std::cout << "Count: " << count << "\n";
            std::cout << "Average Temperature: " << std::fixed << std::setprecision(2) << avg_temperature << "\n";
            std::cout << "Average Humidity: " << std::fixed << std::setprecision(2) << avg_humidity << "\n";
            return 1;
        }
    }
    return 0;
}

int main() {
    
    //Actual input from sensors
    std::string filename = "input/sensor_readings1M.csv";
    std::ifstream file(filename);
    if (!file.is_open()) {
        throw std::runtime_error("Could not open file: " + filename);
    }

    //reference dataset, baseline for filtering
    std::string normal_readings = "input/normal-readings.csv";
    std::ifstream normal_file(normal_readings);
    if (!normal_file.is_open()) {
        throw std::runtime_error("Could not open file: " + normal_readings);
    }

    std::vector<double> temp_data;
    std::vector<double> humidity_data;

    std::string normal_line;
    while (std::getline(normal_file, normal_line)) {
        // Create a stringstream from the line
        std::stringstream ss(normal_line);
        std::string temp, hum;

        // Get the first column value
        if (std::getline(ss, temp, ',')) {
            // Get the second column value
            if (std::getline(ss, hum, ',')) {
                try {
                    // Convert the values to double and add to the vector
                    temp_data.push_back(std::stod(temp));
                    humidity_data.push_back(std::stod(hum));
                } catch (const std::invalid_argument& e) {
                    std::cerr << "Invalid number format: " << temp << " or " << hum << std::endl;                }
            }    
        }
    }

    // Close the file
    normal_file.close();

    int recordsCount = 0;
    std::string line;
    std::unordered_map<int, AggregatedData> sensorDataMap;
    std::unordered_map<int, AggregatedData> agg;
    SensorData filtered;
    auto t_start = std::chrono::high_resolution_clock::now();
    std::unordered_map<int, SensorData> sensorsFiltered;
    
    while (std::getline(file, line)) {
        try {
            auto data = parseCSVLine(line);
            if (filter(data, temp_data, humidity_data)){
                auto& stats = agg[data.sensor_id];
                aggregatedData(stats, data);
            }
            
        } catch (const std::runtime_error& e) {
            std::cerr << "Error processing line: " << e.what() << std::endl;
            continue;
        }
        recordsCount++;
    }

    int total_alerts = 0;
    for (const auto& entry : agg) {
        const auto& sensor_id = entry.first;
        const auto& stats = entry.second;
        double avg_temperature = stats.total_temperature / stats.count;
        double avg_humidity = stats.total_humidity / stats.count;
        if(generateAlert(sensor_id, stats.count, avg_temperature, avg_humidity)==1){
            total_alerts++;
        }
    }   
    
    std::cout << "Total alerts generated: " << total_alerts << "\n";
    auto t_end = std::chrono::high_resolution_clock::now();
    auto time = std::chrono::duration<double>(t_end-t_start).count();
	std::cout << "Execution Time(seconds): " << time << " Throughput (records/s): " << recordsCount/time << std::endl;

    return 0;
}
